/* rnttpd by Vasim V. */
/* $Id: rtnppd.c,v 1.46 1999/11/26 13:09:25 vasim Exp $ */

#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <termios.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <signal.h>
#include <stdarg.h>
#include <string.h>
#include "blowfish.h"
#include "rtnppd.h"
#include "readcfg.h"
#include "crc.h"

#ifdef USE_MSQL
#include "msql.h"
#else
#ifdef USE_POSTGRESQL
#include "libpq-fe.h"
#else
#ifdef USE_MYSQL
#include "mysql.h"
#endif
#endif
#endif

/* Logging levels */
#define L_FATAL LOG_CRIT
#define L_ERROR LOG_ERR
#define L_WARNING LOG_WARNING
#define L_INFO LOG_INFO
#define L_ALL LOG_DEBUG

tnpp_device *head_dev;      /* First TNPP-device */
tnpp_device *tail_dev;      /* Last TNPP-device */
tnpp_device *in_dev;        /* IP-input device */
tnpp_device *ext_dev;       /* External queue */

tnpp_route *head_route;     /* First TNPP-route record */
tnpp_route *tail_route;     /* Last TNPP-route record */
tnpp_device *default_route; /* TNPP-device for default routing */

tnpp_route *head_2route;    /* First backup route record */
tnpp_route *tail_2route;    /* Last backup route record */

tnpp_route *head_3route;    /* First backup (low priority) route record */
tnpp_route *tail_3route;    /* Last backup (low priority) route record */

tnpp_translate *head_trans; /* First TNPP-translate record */
tnpp_translate *tail_trans; /* Last TNPP_translate record */

FILE *err_packs;            /* Output file for bad packets */
FILE *log_packs;            /* Output file for translated packets */

int my_tnpp_addr;           /* TNPP-address for link-test packets */

struct sockaddr_in my_ip_addr; /* IP-address and UDP-port for bind */

char *file_dev_conf;        /* Filename of file with devices config */
char *file_trans_conf;      /* Filename of file with translation tables */
char *file_pid;             /* Filename of file with PID rtnppd */
char *file_main_conf;       /* Filename of main config */
char *section;              /* Section in config files */
char *file_stat;            /* Name of file for dump statistic */
char *file_queue;           /* Name of file for queue packets at exit */

char *u_sock_name;          /* Name of AF_UNIX socket (ext.queue) */
char *u_sock_name_out;      /* Name of AF_UNIX socket (ext.queue out) */
int u_sock;                 /* Ext.queue socket (in) */
struct sockaddr_un sun_out; /* Ext.queue socket destination (out) */
int allowsend;              /* Allow send pages to all local users */

char *sql_serv;             /* Address of SQL-server for logs (may be NULL) */
char *sql_base;             /* Name of SQL-database */
char *sql_port;             /* Port of SQL-server (only for postgresql) */
char *sql_user;             /* SQL-login */
char *sql_passwd;           /* Password for SQL-login */
#ifdef USE_MSQL
int sock_sql;               /* Socket to SQL-server */
#else
#ifdef USE_POSTGRESQL
PGconn *conn_sql;
#else
#ifdef USE_MYSQL
MYSQL *conn_sql;
#endif
#endif
#endif

/* Global flags */
int flag_dec_inertia;         /* if 1 - must decrement inertia as in TNPP
			       * specification */
int flag_fatal_error;         /* Fatal error during loop (e.g. can't open
			       * serial device, etc) */
int flag_reread;              /* Reread config files */
int loglevel;                 /* Level of logging */
int log_stderr;               /* Use stderr instead syslog */
int one_line_log;             /* Log packets in one line */
int use_sql_log;              /* Log packets in SQL-database */
int ext_queue;                /* Use external queue-utility */
int ext_rdonly;

int timeout_next;             /* Timeout to reroute packet via another route */

/* Default timers */
int i_tidle;
int i_tidle_enq;
int i_tict;
int i_tnri;
int i_tictout;
int i_cretry;
int i_chold;
int i_cenq;
int i_thold;
int max_packs;

void nonblock(int fd);
void reconnect_serial(tnpp_device *dev);      /* Pulse DTR, wait carrier */
void tnpp_init(tnpp_device *dev); /* send ENQ, wait EOT, send empty packet */
void log_tnpp_packet(FILE *out, unsigned char *data,
		     int transparentcrc, char *error); /* Log packet in file */
int tnpp_checksum(unsigned char *data, int transparentcrc);
                              /* Control checksum (0 - bad) */
int process_tnpp_packet(tnpp_device *indev); /* Route, translate and queue
					      * packet from device buffer
					      * Return 0, if errors */
void time_packet(tnpp_packet *pack);         /* Show time of processing */
void send_empty_packet(tnpp_device *dev);    /* Put header-only packet
					      * (link test) */
void queue_packet(tnpp_device *dev, tnpp_packet *pack); /* Put packet in
							 * queue (tail) */
tnpp_packet *create_packet(); /* Create empty packet */
tnpp_packet *create_ack(unsigned char ack, tnpp_device *dev);
void queue_ack(tnpp_device *dev, unsigned char ack); /* Queue one-byte
						      * packet */
void dequeue_head(tnpp_device *dev); /* Remove first packet from queue */
void queue_first(tnpp_device *dev, tnpp_packet *pack); /* Put packet to
							* head of queue */
void wlog(int prio, char *fmt, ...); /* Log message */
int reread_device_conf();            /* Read devices config
				      * return 0, if failed */
int reread_trans_conf();             /* Read translation config
				      * return 0, if failed */
int dest_packet(tnpp_packet *pack);  /* Return destination of packet */
int translate_packet(tnpp_packet *pack, int transparentcrc);
                                     /* Translate destination address and
				      * log packet (if need)
				      * Return 1, if was translated */
int encrypt_packet(tnpp_packet *pack, tnpp_device *dev); /* Encrypt packet */
void send_page(tnpp_device *dev, int message);

/* Put packet into file */
void put_to_file(char *filename, unsigned char *data, int tcrc) {
  int len;
  FILE *f;
  unsigned char *p;
  unsigned char buf[5];
  int crc;
  int i;

  for(len = 0; data[len] != ETX; len++);
  len++;
  p = data + len;
  if (tcrc) {
    for(i = 0; i < 4; i++)
      buf[i] = toupper(p[i+1]);
    buf[4] = '\0';
    crc = strtol(buf, NULL, 16);
  } else
    crc = *((unsigned char *) (p + 1)) + 256 * (*((unsigned char *) (p+2)));

  data[len] = crc & 0xff;
  data[len + 1] = (crc / 256) & 0xff;
  len = len + 2;
  f = fopen(filename, "a");
  if (!f) {
    wlog(L_ERROR, "Can't open file '%s' (%s)! Packet lost",
	 filename, strerror(errno));
    return;
  }
  fprintf(f, "%d\n", len);
  fwrite(data, len, 1, f);
  fclose(f);
} /* put_to_file */

/* Translate, route and queue TNPP-packet */
/* Return 0, if errors */
int route_tnpp_packet(tnpp_device *indev, unsigned char *pdata, int flag) {
  int len;
  tnpp_packet *pack;
  tnpp_device *dev;
  tnpp_route *route;
  int inertia, crc, dest, serial, i, src;
  char buf[5];
  unsigned char *data;

  data = pdata;
  pack = create_packet();
  if (!pack)
    return 0;
  for(len = 0; data[len] != ETX; len++);
  if (indev->transparentcrc)
    len = len + 5;
  else
    len = len + 3;
  pack->length = len;
  if (len < 13) {
    free(pack);
    indev->pktinerr++;
    return 0;
  }
  memcpy(pack->bytes, data, len);
  gettimeofday(&pack->begintime, NULL);

  /* Translate packet (if need) */
  translate_packet(pack, indev->transparentcrc);

  /* Decrement inertia */
  if (flag_dec_inertia && (indev != ext_dev)) {
    buf[0] = toupper(pack->bytes[5]);
    buf[1] = toupper(pack->bytes[6]);
    buf[2] = '\0';
    inertia = strtol(buf, NULL, 16) - 1;
    if (inertia < 0) {
      wlog(L_WARNING, "Inertia for packet is zero. Packet dropped");
      log_tnpp_packet(err_packs, pack->bytes,
		      indev->transparentcrc, "inertia == 0");
      time_packet(pack);
      free(pack);
      indev->pktinerr++;
      return 0;
    }
    snprintf(buf, 3, "%02X", inertia);
    memcpy(pack->bytes + 5, buf, 2);
  }

  /* Checking for zero or duplicated packet */
  if (indev != ext_dev) {
    buf[0] = toupper(pack->bytes[11]);
    buf[1] = toupper(pack->bytes[12]);
    buf[2] = '\0';
    serial = strtol(buf, NULL, 16);
    if (!serial) {
      wlog(L_INFO, "Zero serial. Resetting serial. Packet dropped");
      for(i = 0; i < SERIAL_TABLE_NUM; i++)
	indev->serials[i] = 0;
      indev->serials_end = 0;
      time_packet(pack);
      free(pack);
      return 1;
    } else {
      for(i = 0; i < SERIAL_TABLE_NUM; i++)
	if (serial == indev->serials[i]) {
	  wlog(L_INFO, "Duplicated packet. Packet dropped");
	  time_packet(pack);
	  indev->pktinerr++;
	  free(pack);
	  return 1;
	}
      indev->serials[indev->serials_end] = serial;
      indev->serials_end = (indev->serials_end + 1) % SERIAL_TABLE_NUM;
    }
  }

  /* Checking for zero destination */
  dest = dest_packet(pack);
  if (dest == 0) {
    wlog(L_INFO, "Zero destination. Packet dropped");
    time_packet(pack);
    indev->pktinerr++;
    free(pack);
    return 1;
  }

  /* Send to external queue */
  if (!flag && ext_queue && (indev != ext_dev)) {
    dev = ext_dev;
    goto ROUTE;
  }

  if (flag) {
    /* Search for backup route */
    for (route = head_2route; route; route = route->next) {
      if ((dest <= route->high_addr) && (dest >= route->low_addr)) {
	dev = route->dev;
	if (dev != indev) {
	  wlog(L_INFO, "Using backup (2) route for %04X (device '%s')",
	       dest, dev->name);
	  goto ROUTE;
	}
	break;
      }
    }
    for (route = head_3route; route; route = route->next) {
      if ((dest <= route->high_addr) && (dest >= route->low_addr)) {
	dev = route->dev;
	if (dev != indev) {
	  wlog(L_INFO, "Using backup (3) route for %04X (device '%s')",
	       dest, dev->name);
	  goto ROUTE;
	}
	break;
      }
    }
  }

  /* Search direct route */
  for(dev = head_dev; dev; dev = dev->next)
    if (dev->addr == dest)
      goto ROUTE;

  /* Search indirect route */
  for(route = head_route; route; route = route->next)
    if ((dest <= route->high_addr) && (dest >= route->low_addr)) {
      dev = route->dev;
      goto ROUTE;
    }

  /* Try use default route */
  if (default_route)
    dev = default_route;
  else {
    wlog(L_WARNING, "No route for %04X. Packet dropped",
	 dest);
    log_tnpp_packet(err_packs, pack->bytes, indev->transparentcrc, "no route");
    time_packet(pack);
    free(pack);
    indev->pktinerr++;
    return 0;
  }

  if (dev->num_packs >= max_packs) {
    wlog(L_WARNING, "No queue space for device '%s'", dev->name);
    log_tnpp_packet(err_packs, pack->bytes, indev->transparentcrc, "no space");
    put_to_file(file_queue, pack->bytes, indev->transparentcrc);
    time_packet(pack);
    indev->pktin++;
    free(pack);
    return 0;
  }

  /* Ok, put packet in device's queue */
 ROUTE:
  indev->pktin++;
  wlog(L_INFO, "Packet for %04X was routed in device '%s'",
       dest, dev->name);
  /* Change zero source to our address, if packet from ext_dev */
  if (indev == ext_dev) {
    for(i = 0; i < 4; i++)
      buf[i] = toupper(pack->bytes[i + 7]);
    buf[4] = '\0';
    src = strtol(buf, NULL, 16);
    if (!src) {
      sprintf(buf, "%04X", my_tnpp_addr);
      memcpy(pack->bytes + 7, buf, 4);
    }
  }
  /* New serial */
  dev->last_serial++;
  if (dev->last_serial > 255)
    dev->last_serial = 1;
  sprintf(buf, "%02X", dev->last_serial);
  memcpy(pack->bytes + 11, buf, 2);
  /* Calculate new crc */
  if (indev->transparentcrc)
    len = len - 4;
  else
    len = len - 2;
  crc = crc16(pack->bytes, len);
  if (dev->transparentcrc) {
    if ((len + 4) > MAX_PACKET_SIZE) {
      wlog(L_WARNING, "No buffer space in device '%s' (no space for T-CRC)",
	   dev->name);
      log_tnpp_packet(err_packs, pack->bytes, indev->transparentcrc,
		      "no space for T-CRC");
      put_to_file(file_queue, pack->bytes, indev->transparentcrc);
      time_packet(pack);
      free(pack);
      return 0;
    }
    snprintf(buf, sizeof(buf), "%04X", crc);
    memcpy(pack->bytes + len, buf, 4);
    len = len + 4;
  } else {
    pack->bytes[len] = crc & 0xff;
    pack->bytes[len + 1] = (crc / 256) & 0xff;
    len = len + 2;
  }
  pack->length = len;
  /* Queue packet */
  pack->status = 0;
  pack->offset = 0;
  pack->thold = i_thold;
  pack->chold = i_chold;
  queue_packet(dev, pack);
  return 1;
} /* route_tnpp_packet */

int process_tnpp_packet(tnpp_device *indev) {
  return route_tnpp_packet(indev, indev->bytes, 0);
} /* process_tnpp_packet */

void reroute_head(tnpp_device *dev) {
  unsigned char buf[MAX_PACKET_SIZE];

  memcpy(buf, dev->head->bytes, dev->head->length);
  if (dev->encrypt)
    decrypt_packet(buf, dev);
  dequeue_head(dev);
  route_tnpp_packet(dev, buf, 1);
} /* reroute_head */

/* Encrypt packet */
int encrypt_packet(tnpp_packet *pack, tnpp_device *dev) {
  int len, lencr;
  unsigned char buf[MAX_PACKET_SIZE];
  unsigned char bufcrc[5];
  u_int64_t init;
  int i, j;
  unsigned char c;

  len = 1;
  while (pack->bytes[len] != ETX) len++;
  len--;
  if (len <= 0)
    return 0;
  lencr = (len + 7) & 0xfff8;
  if ((lencr + 8) > MAX_PACKET_SIZE)
    return 0;
  memcpy(bufcrc, pack->bytes + len + 1, (dev->transparentcrc) ? 5 : 3);
  /* Align data */
  if (lencr > len) {
    for(i = (len + 1); i < (lencr + 1); i++)
      pack->bytes[i] = random() & 0xff;
  }
  /* Encrypt packet and create subheader */
  bf_encrypt_data(pack->bytes + 1, lencr, ((u_int64_t) 0xA5A5A5A5 << 32)
		  + (u_int64_t) 0x5A5A5A5A, &dev->bf->skeys);
  buf[0] = pack->bytes[0];
  snprintf(buf + 1, 5, "%04X", len);
  j = 5;
  /* Create transparent-packet */
  for(i = 0; i < lencr; i++) {
    c = pack->bytes[i + 1];
    if ((c == SOH) || (c == ETX) || (c == ENQ) || (c == EOT) || (c == SUB)
	|| (c == NAK) || (c == RS) || (c == CAN) || (c == ACK)) {
      buf[j] = SUB;
      j++;
      buf[j] = (c + 0x40) & 0xff;
    } else
      buf[j] = c;
    j++;
    if (j >= MAX_PACKET_SIZE)
      return 0;
  }
  if (dev->transparentcrc)
    len = 5;
  else
    len = 3;
  if (j + len >= MAX_PACKET_SIZE)
    return 0;
  memcpy(buf + j, bufcrc, len);
  pack->length = j + len + 1;
  memcpy(pack->bytes, buf, pack->length);
  pack->encrypt = 1;
  return 1;
} /* encrypt_packet */

/* Create page message from string '<tnpp-addr>:<text-from-config>' */
void send_page(tnpp_device *dev, int message) {
  char *p;
  int addr;
  int i;
  int len;
  char buf[5];
  int crc;
  char *msg;

  if ((!dev->page) || (dev->page_already == message))
    return;
  dev->page_already = message;
  switch (message) {
  case 0:
    msg = "LINK DOWN";
    break;
  case 1:
    msg = "LINK UP";
    break;
  case 2:
    msg = "LINK ERR";
    break;
  }
  p = strchr(dev->page, ':') + 1;
  for (i = 0; (i < 4) && (dev->page[i] != ':'); i++)
    buf[i] = toupper(dev->page[i]);
  buf[4] = '\0';
  addr = strtol(buf, NULL, 16);
  in_dev->last_serial++;
  if (in_dev->last_serial >= 256)
    in_dev->last_serial = 1;
  sprintf(in_dev->bytes, "%c%04X20%04X%02X%c", SOH, addr, my_tnpp_addr,
	   in_dev->last_serial, STX);
  snprintf(in_dev->bytes + strlen(in_dev->bytes), 900, p, msg);
  len = strlen(in_dev->bytes);
  in_dev->bytes[len] = ETX;
  len++;
  crc = crc16(in_dev->bytes, len);
  in_dev->bytes[len] = crc & 0xff;
  in_dev->bytes[len + 1] = (crc / 256) & 0xff;
  if (!process_tnpp_packet(in_dev))
    wlog(L_WARNING, "Could not send fault page for device '%s'", dev->name);
} /* send_page */

/* Decrypt packet */
int decrypt_packet(unsigned char *data, tnpp_device *dev) {
  int len, lencr, lencrc;
  unsigned char buf[MAX_PACKET_SIZE];
  int i, j, flag_sub;
  unsigned char c;

  memcpy(buf, data + 1, 4);
  buf[4] = '\0';
  len = strtol(buf, NULL, 16);
  if (len > MAX_PACKET_SIZE)
    return 0;
  buf[0] = data[0];
  flag_sub = 0;
  j = 1;
  for(i = 5; data[i] != ETX; i++) {
    c = data[i];
    if (flag_sub) {
      if (c < 0x40)
	c = c + 0xc0;
      else
	c = c - 0x40;
      buf[j] = c;
      j++;
      flag_sub = 0;
    } else
      if (c != SUB) {
	buf[j] = c;
	j++;
      } else
	flag_sub = 1;
  }
  lencr = j - 1;
  if ((len > lencr) || (len < (lencr - 7)))
    return 0;
  bf_decrypt_data(buf + 1, lencr, ((u_int64_t) 0xA5A5A5A5 << 32)
		  + (u_int64_t) 0x5A5A5A5A, &dev->bf->skeys);
  if (dev->transparentcrc)
    lencrc = 5;
  else
    lencrc = 3;
  memcpy(buf + len + 1, data + i, lencrc);
  len = len + lencrc + 1;
  memcpy(data, buf, len);
  if (!tnpp_checksum(buf, dev->transparentcrc))
    return 0;
  return 1;
} /* decrypt_packet */

/* Return destination of packet */
int dest_packet(tnpp_packet *pack) {
  char buf[5];
  int i;

  for(i = 0; i < 4; i++)
    buf[i] = toupper(pack->bytes[i + 1]);
  buf[4] = '\0';
  return strtol(buf, NULL, 16);
} /* dest_packet */

/* Translate and log packet */
int translate_packet(tnpp_packet *pack, int transparentcrc) {
  tnpp_translate *trans;
  int dest;
  unsigned char buf[9];

  dest = dest_packet(pack);
  for(trans = head_trans; trans; trans = trans->next)
    if (dest == trans->old_addr) {
      wlog(L_INFO, "Translation: old destination %04X, new destination %04X",
	   trans->old_addr, trans->new_addr);
      if (dest != trans->new_addr) {
	snprintf(buf, sizeof(buf), "%04X", trans->new_addr);
	memcpy(pack->bytes + 1, buf, 4);
      }
      if (trans->logging)
	log_tnpp_packet(log_packs, pack->bytes, transparentcrc, "log");
      if (trans->cap[0].change != 0) {
	/* Perform CAP function translation */
	int offs;
	unsigned long code;
	char func;
	int i;

	if (pack->bytes[14] == 0x3E)
	  offs = 17;
	else
	  offs = 14;
	if (pack->bytes[offs] == 'A') {
	  buf[8] = '\0';
	  memcpy(buf, pack->bytes + offs + 6, 8);
	  code = strtoul(buf, NULL, 16);
	  for (i = 0; i < CAP_TRANS_NUM; i++) {
	    if (trans->cap[i].change == 0)
	      break;
	    if ((code >= trans->cap[i].low) && (code <= trans->cap[i].high)) {
	      func = pack->bytes[offs + 5];
	      pack->bytes[offs + 5] = func + trans->cap[i].change;
	      wlog(L_INFO, "Translation: CAP %08X, old func '%c', new '%c'",
		   code, func, func, pack->bytes[offs + 5]);
	    }
	  }
	}
      }
      return 1;
    }
  return 0;
} /* translate_packet */

/* Connect to serial port and toggle DTR */
void reconnect_serial(tnpp_device *dev) {
  struct termios term;
  struct timeval tm;
  int n;
  const struct {
    int baud;
    speed_t t_baud;
  } speeds[7] = {{300, B300}, {600, B600}, {1200, B1200}, {2400, B2400},
		 {4800, B4800}, {9600, B9600}, {9600,B9600}};

  if (dev->type != 0)
    return;
  if (!(dev->fd < 0))
    close(dev->fd);
  dev->fd = open(dev->device, O_RDWR | O_NONBLOCK);
  if (dev->fd < 0) {
    wlog(L_ERROR, "Error %s (%d) at open '%s' for device '%s'",
	 strerror(errno), errno, dev->device, dev->name);
    flag_fatal_error = 1;
    return;
  }
  tcgetattr(dev->fd, &term);
  for(n = 0; n < 6; n++)
    if (speeds[n].baud == dev->serial_speed)
      break;
  cfsetispeed(&term, speeds[n].t_baud);
  cfsetospeed(&term, speeds[n].t_baud);
  term.c_cflag &= ~(CSTOPB | PARENB | PARODD | HUPCL | CLOCAL | CRTSCTS
		    | CSIZE); 
  term.c_cflag |= CS8 | CREAD;
  if (dev->serial_crtscts)
    term.c_cflag |= CRTSCTS;
  term.c_cflag |= CLOCAL;
  term.c_iflag = 0;
  term.c_oflag = 0;
  term.c_lflag = 0;
  term.c_cc[VMIN] = 1;
  term.c_cc[VTIME] = 0;
  if (tcsetattr(dev->fd, TCSANOW, &term) < 0) {
    wlog(L_ERROR, "Error %s (%d) at set serial parameters for device '%s'",
	 strerror(errno), errno, dev->name);
    flag_fatal_error = 1;
    return;
  }
  nonblock(dev->fd);
#ifdef TIOCM_DTR
  if (dev->serial_dtrtoggle) {
    n = TIOCM_DTR;
    /* Toggle DTR */
    ioctl(dev->fd, TIOCMBIC, &n);
    tm.tv_sec = 0;
    tm.tv_usec = 250000;
    select(0, NULL, NULL, NULL, &tm);
    ioctl(dev->fd, TIOCMBIS, &n);
    tm.tv_sec = 0;
    tm.tv_usec = 250000;
    select(0, NULL, NULL, NULL, &tm);
  } else {
    n = TIOCM_DTR;
    ioctl(dev->fd, TIOCMBIS, &n);
  }
#endif
} /* reconnect_serial */

/* Log string via syslog */
void wlog(int prio, char *fmt,...) {
  va_list ap;
  char buf[128];
  time_t ltime;
  char *timestamp;
  char buftime[16];

  if (loglevel >= prio) {
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);
    if (log_stderr) {
      time(&ltime);
      timestamp = ctime(&ltime);
      memcpy(buftime, timestamp + 4, 15);
      buftime[15] = '\0';
      fprintf(stderr, "%s %s\n", buftime, buf);
    } else
      syslog(prio, buf);
  }
} /* wlog */

/* Put empty header-only packet in queue (after link test) */
void send_empty_packet(tnpp_device *dev) {
  tnpp_packet *pack;
  unsigned short int crc;

  pack = create_packet();
  if (pack) {
    pack->type = 1;
    gettimeofday(&pack->begintime, NULL);
    snprintf(pack->bytes, MAX_PACKET_SIZE, "%c000001%04X00%c%c",
	     SOH, my_tnpp_addr, STX, ETX);
    pack->length = strlen(pack->bytes);
    crc = crc16(pack->bytes, pack->length);
    if (dev->transparentcrc) {
      snprintf(pack->bytes + pack->length, MAX_PACKET_SIZE - pack->length,
	       "%04X", crc);
      pack->length = pack->length + 4;
    } else {
      pack->bytes[pack->length] = (unsigned char) (crc & 0xff);
      pack->bytes[pack->length + 1] = (unsigned char) ((crc / 256) & 0xff);
      pack->length = pack->length + 2;
    }
    queue_packet(dev, pack);
  }
} /* send_empty_packet */

/* Put packet in tail of queue */
void queue_packet(tnpp_device *dev, tnpp_packet *pack) {
  pack->prev = dev->tail;
  if (dev->tail)
    dev->tail->next = pack;
  dev->tail = pack;
  dev->num_packs++;
  if (!dev->head) {
    dev->head = pack;
    if (dev->statusout == 0) {
      dev->statusout = 1;
      dev->tictout = i_tictout;
      dev->cretry = i_cretry;
      dev->tnri = i_tnri;
    }
  }
} /* queue+packet */

/* Put packet in head of queue (do not change statusout and num_packs) */
void queue_first(tnpp_device *dev, tnpp_packet *pack) {
  if (pack) {
    pack->next = dev->head;
    if (dev->head)
      dev->head->prev = pack;
    if (!dev->tail)
      dev->tail = pack;
    dev->head = pack;
  }
} /* queue_first */ 
    
/* Send ENQ, wait EOT and send empty packet after */
void tnpp_init(tnpp_device *dev) {
  int i;

  for(i = 0; i < SERIAL_TABLE_NUM; i++)
    dev->serials[i] = 0;
  dev->serials_end = 0;
  dev->last_serial = 0;
  dev->status = 5;
  if ((dev->statusout == 1) && dev->head) {
    dev->head->offset = 0;
    dev->head->status = 0;
  }
  dev->statusout = 2;
  dev->offset = 0;
  dev->tidle = 0;
} /* tnpp_init */

/* Check CRC in packet */
/* Return 0, if CRC bad */
int tnpp_checksum(unsigned char *data, int transparentcrc) {
  unsigned char *p;
  int length, i;
  unsigned char buf[5];
  unsigned short int crc;

  for(p = data; *p != ETX; p++);
  length = p - data + 1;
  if (transparentcrc) {
    for(i = 0; i < 4; i++)
      buf[i] = toupper(p[i+1]);
    buf[4] = '\0';
    crc = strtol(buf, NULL, 16);
  } else
    crc = *((unsigned char *) (p + 1)) + 256 * (*((unsigned char *) (p+2)));
  if (crc != crc16(data, length))
    return 0;
  else
    return 1;
} /* tnpp_checksum */

/* Create packet with empty data */
tnpp_packet *create_packet() {
  tnpp_packet *pack;

  pack = (tnpp_packet *) malloc(sizeof(tnpp_packet));
  if (pack) {
    pack->prev = NULL;
    pack->next = NULL;
    pack->status = 0;
    pack->thold = i_thold;
    pack->chold = i_chold;
    pack->offset = 0;
    pack->length = 0;
    pack->encrypt = 0;
    pack->begintime.tv_sec = 0;
    pack->begintime.tv_usec = 0;
    pack->type = 0;
  }
  return pack;
} /* create_packet */

/* Create one-byte packet */
tnpp_packet *create_ack(unsigned char ack, tnpp_device *dev) {
  tnpp_packet *pack;
  unsigned short int crc;
  long rn;
  unsigned char buf[9];
  int i, j;

  pack = create_packet();
  if (pack) {
    pack->type = 1;
    pack->bytes[0] = ack;
    if (dev->encrypt) {
      /* Create ack with encrypted signature */
      pack->encrypt = 1;
      buf[0] = ack;
      for(i = 1; i < 7; i += 3) {
	rn = random();
	buf[i+2] = (rn >> 16) & 0xff;
	buf[i+1] = (rn >> 8) & 0xff;
	buf[i] = rn & 0xff;
      }
      crc = crc16(buf, 7);
      buf[7] = crc & 0xff;
      buf[8] = (crc >> 8) & 0xff;
      j = 1;
      bf_encrypt_data(buf + 1, 8, 0, &dev->bf->skeys);
      for(i = 1; i < 9; i++) {
	snprintf(pack->bytes + j, 3, "%02X", buf[i]);
	j = j + 2;
      }
      pack->length = 17;
    } else
      pack->length = 1;
    gettimeofday(&pack->begintime, NULL);
  }
  return pack;
} /* create_ack */

/* Decrypt encrypted ack */
int decrypt_ack(unsigned char *data, tnpp_device *dev) {
  unsigned short int crc;
  int i, j;
  unsigned char buf[9];
  unsigned char hd[3];

  j = 1;
  for(i = 1; i < 17; i += 2) {
    hd[0] = data[i];
    hd[1] = data[i + 1];
    hd[2] = '\0';
    buf[j] = strtol(hd, NULL, 16);
    j++;
  }
  bf_decrypt_data(buf + 1, 8, 0, &dev->bf->skeys);
  buf[0] = data[0];
  crc = (unsigned short int) buf[7] + ((unsigned short int) buf[8] << 8);
  if (crc != crc16(buf, 7)) {
    wlog(L_WARNING, "Bad CRC in encrypted ACK at device '%s'",
	 dev->name);
    return 0;
  }
  return 1;
} /* decrypt_ack */

/* Put one-byte packet in queue */
void queue_ack(tnpp_device *dev, unsigned char ack) {
  tnpp_packet *pack;

  if (dev->num_packs >= max_packs)
    return;
  pack = create_ack(ack, dev);
  if (pack) {
    dev->num_packs++;
    if (dev->head) {
      if ((dev->head->status == 2) && (dev->statusout == 0))
	queue_first(dev, pack);
      else {
	/* If there others packet in queue, put ack after first packet */
	pack->prev = dev->head;
	pack->next = dev->head->next;
	if (pack->next)
	  pack->next->prev = pack;
	dev->head->next = pack;
	if (pack->next == NULL)
	  dev->tail = pack;
      }
    } else {
      dev->head = pack;
      dev->tail = pack;
      if (dev->statusout == 0) {
	dev->statusout = 1;
	dev->tictout = i_tictout;
	dev->cretry = i_cretry;
	dev->tnri = i_tnri;
      }
    }
  }
} /* queue_ack */

/* Remove packet from head of queue */
void dequeue_head(tnpp_device *dev) {
  tnpp_packet *pack;

  dev->tnri = i_tnri;
  dev->cretry = i_cretry;
  /* Remove packet from queue */
  pack = dev->head;
  if (pack) {
    dev->num_packs--;
    dev->head = pack->next;
    if (dev->tail == pack)
      dev->tail = NULL;
    if (dev->head) {
      dev->head->prev = NULL;
      if ((dev->statusout == 0) || (dev->statusout == 1)) {
	if (dev->head->status == 2)
	  dev->statusout = 0;
	else {
	  dev->head->status = 1;
	  dev->statusout = 1;
	}
      }
    } else
      if (dev->statusout == 1)
	dev->statusout = 0;
    time_packet(pack);
    free(pack);
  }
} /* dequeue_head */

void time_packet(tnpp_packet *pack) {
  if (!pack)
    return;
  if ((pack->begintime.tv_sec != 0) || (pack->begintime.tv_usec != 0)) {
    struct timeval endtime;
    int sec, msec;
    
    gettimeofday(&endtime, NULL);
    sec = endtime.tv_sec - pack->begintime.tv_sec
      + (endtime.tv_usec/1000000 - pack->begintime.tv_usec/1000000);
    msec = (endtime.tv_usec/1000 - pack->begintime.tv_usec/1000) % 1000;
    if (msec < 0)
      msec = 1000 + msec;
    wlog(L_INFO, "Packet was processed in %d.%03d seconds", sec, msec);
  }
} /* time_packet */


/* Main loop */
int loop() {
  tnpp_packet *pack;
  tnpp_device *dev;
  tnpp_device *dev2;
  fd_set rfd, wfd, efd;
  struct timeval tm;
  int res, nfds;
  long t_bef, t_aft;
  int len;
  struct sockaddr_in peer;
  unsigned char c;
  int n, offset;
  int crcok;
  char *p;

  flag_reread = 0;
  flag_fatal_error = 0;
  wlog(L_WARNING, "Entering into main loop");

  gettimeofday(&tm, NULL);
  t_bef = tm.tv_sec;

  /* Run to break */
  while (!flag_fatal_error && !flag_reread) {
    /* Initializing descriptor table for select */
    FD_ZERO(&rfd);
    FD_ZERO(&wfd);
    FD_ZERO(&efd);
    nfds = 0;
    for (dev = head_dev; dev; dev = dev->next) {
      if ((dev->status != 255)  /* We wait for receive */
	  && ((dev->type != 1) || (dev == in_dev))) /* only on serial and
						     * IP-input devices */
	FD_SET(dev->fd, &rfd);
      if ((dev->statusout == 1) /* We wait for send */
	  || ((dev->statusout == 2) && (dev->tidle < 0))
	  || (dev->statusout == 3))
	FD_SET(dev->fd, &wfd);
      FD_SET(dev->fd, &efd);    /* We wait... Uh, don't know... */
      if (nfds < dev->fd)
	nfds = dev->fd;
    }

    /* Only one second wait (then change timers, etc...) */
    tm.tv_sec = 1;
    tm.tv_usec = 0;

    wlog(L_ALL, "Entering in select with max fd %d", nfds);
    res = select(nfds+1, &rfd, &wfd, &efd, &tm);
    if (res == -1) {
      if (errno != EINTR) {
	wlog(L_FATAL,
	     "Error '%s' (%d) at select. Exiting", strerror(errno), errno);
	break;
      }
      wlog(L_ALL, "EINTR during select");
      FD_ZERO(&rfd);
      FD_ZERO(&efd);
      FD_ZERO(&wfd);
    }
    wlog(L_ALL, "select return %d ready sockets", res);
    gettimeofday(&tm, NULL);
    t_aft = tm.tv_sec;

    /* Checking all devices for ready sockets */
    for (dev = head_dev; dev; dev = dev->next) {
      if (FD_ISSET(dev->fd, &rfd)) {
	/* Receive some data (for socket - with recvfrom) */
	if (dev->type == 0)
	  res = read(dev->fd, &dev->bytes[dev->offset],
		     MAX_PACKET_SIZE-dev->offset);
	else {
	  len = sizeof(peer);
	  res = recvfrom(dev->fd, &dev->bytes[dev->offset],
			 MAX_PACKET_SIZE-dev->offset, 0,
			 (struct sockaddr *) &peer, &len);
	}
	/* Checking for errors */
	if (res < 0) {
	  /* Errors at read */
	  if ((errno != EINTR) && (errno != EAGAIN)) {
	    wlog(L_ERROR, "Error '%s' (%d) at read from device '%s'",
		 strerror(errno), errno, dev->name);
	    if ((dev->type == 0) && (dev != ext_dev)) {
	      /* Trying reconnect serial device */
	      reconnect_serial(dev);
	      tnpp_init(dev);
	    }
	    if ((dev != ext_dev) && (dev != in_dev))
	      send_page(dev, 2);
	  } else
	    wlog(L_ALL, "EINTR or EAGAIN received at read from device '%s'",
		 dev->name);
	} else {
	  /* read without errors */
	  wlog(L_ALL, "%d bytes was received from device '%s'",
	       res, dev->name);
	  if (dev->type == 1) {
	    /* Socket devices have only one receiver
	     * We must find right device for new data */
	    for (dev2 = head_dev; dev2; dev2 = dev2->next) {
	      if ((dev2->peer.sin_addr.s_addr == peer.sin_addr.s_addr)
		  && (dev2->peer.sin_port == peer.sin_port)
		  && (dev2->peer.sin_family == peer.sin_family)) {
		wlog(L_ALL, "New bytes moved in device '%s' buffer",
		     dev2->name);
		if (res > (MAX_PACKET_SIZE - dev2->offset)) {
		  wlog(L_WARNING, "No buffer space for device '%s'",
		       dev2->name);
		  res = MAX_PACKET_SIZE - dev2->offset;
		}
		memcpy(&dev2->bytes[dev2->offset],
		       &dev->bytes[dev->offset], res);
		goto FOUND;
	      }
	    }
	    wlog(L_WARNING, "Packet from unknown host %s:%d",
		 inet_ntoa(peer.sin_addr), htons(peer.sin_port));
	    res = 0;
	    dev2 = dev;
	  } else
	    dev2 = dev;
	FOUND:
	  if (res > 0) {
	    /* New bytes! Reseting idle timer */
	    dev2->tidle = i_tidle;
	    dev2->tict = i_tict;
	    offset = dev2->offset;
	    dev2->offset = dev2->offset + res;
	    /* Bytes arrived, will check for control characters */
	    for (n = 0; n < res; n++) {
	      c = dev2->bytes[offset + n];

	      if ((dev2->status == 2) || (dev2->status == 3)
		  || (dev2->status == 6) || (dev2->status == 7)) {
		/* checksum's bytes */
		if (dev2->status == 3) {
		  /* Receive done! Control checksum and route it */
		  if (dev2->encrypt)
		    crcok = decrypt_packet(dev2->bytes, dev2);
		  else
		    crcok = tnpp_checksum(dev2->bytes, dev2->transparentcrc);
		  if (crcok) {
		    /* Checksum correct */
		    wlog(L_INFO, "Route new packet from device '%s'",
			 dev2->name);
		    if (dev2 != ext_dev) {
		      if (process_tnpp_packet(dev2))
			queue_ack(dev2, ACK);
		      else
			queue_ack(dev2, CAN);
		    } else
		      (void) process_tnpp_packet(dev2);
		  } else {
		    /* Checksum error. Queue NAK after current packet */
		    wlog(L_INFO, "Bad checksum in packet at device '%s'",
			 dev2->name);
		    dev2->pktinerr++;
		    log_tnpp_packet(err_packs, dev2->bytes,
				    dev->transparentcrc,"bad checksum");
		    if (dev2 != ext_dev)
		      queue_ack(dev2, NAK);
		  }
		  if (((dev2->statusout == 0) || (dev2->statusout == 2)) &&
		      dev2->head)
		    dev2->statusout = 1;
		  if (n != (res - 1)) {
		    /* There some bytes in buffer... */
		    memmove(dev2->bytes, &dev2->bytes[offset+n], res-n);
		    res = res - n;
		    dev2->offset = dev2->offset - (n + offset);
		    n = 0;
		    offset = 0;
		  } else
		    dev2->offset = 0;
		  dev2->status = 0;
		}

		if (dev2->status == 2) {
		  /* Wait last byte in checksum */
		  if (dev2->transparentcrc)
		    dev2->status = 6;
		  else
		    dev2->status = 3;
		} else
		  if (dev2->status == 6)
		    dev2->status = 7;
		  else
		    if (dev2->status == 7)
		      dev2->status = 3;
	      } else
		switch (c) {
		case SOH:
		  /* New packet */
		  if ((n > 0) || (offset > 0)) {
		    wlog(L_INFO, "New packet before ETX at device '%s'",
			 dev2->name);
		    memmove(dev2->bytes, &dev2->bytes[offset+n], res-n);
		    dev2->offset = dev2->offset - (n + offset);
		    res = res - n;
		    n = 0;
		    offset = 0;
		    dev2->pktinerr++;
		  }
		  if ((dev2->status == 4) || (dev2->status == 5))
		    send_page(dev2, 1);
		  dev2->status = 1;
		  break;

		case ETX:
		  /* End packet. Will wait CRC */
		  if (dev2->status == 1)
		    dev2->status = 2;
		  break;

		case EOT:
		  /* EOT. If we wait it, up line. If no - initialize */
		  if (dev2->encrypt) {
		    if ((res - n) < 17)
		      break;
		    if (!decrypt_ack(dev->bytes + offset + n, dev2))
		      break;
		    n = n + 16;
		  }
		  if (dev2->head) {
		    dev2->head->offset = 0;
		    if (dev2->head->type != 1)
		      dev2->pktouterr++;
		  }
		  if (dev2->statusout == 3) {
		    queue_ack(dev2, EOT);
		    dev2->statusout = 1;
		  }
		  if (dev2->status == 5) {
		    send_empty_packet(dev2);
		  } else
		    if (dev2->status != 4) {
		      wlog(L_INFO, "Unexpected EOT at device '%s'",
			   dev2->name);
		      break;
		    }

		  wlog(L_WARNING, "EOT received at device '%s'. Line up",
		       dev2->name);
		  dev2->status = 0;
		  if (dev2->statusout == 2) {
		    /* Stop link test */
		    if (dev2->head) {
		      dev2->statusout = 1;
		      dev2->head->status = 1;
		    } else
		      dev2->statusout = 0;
		  }
		  dev2->tidle = i_tidle;
		  dev2->cenq = i_cenq;
		  send_page(dev2, 1);
		  break;

		case ENQ:
		  /* Link test. Send EOT */
		  if (dev2->encrypt) {
		    if ((res - n) < 17)
		      break;
		    if (!decrypt_ack(dev->bytes + offset + n, dev2))
		      break;
		    n = n + 16;
		  }
		  wlog(L_INFO, "ENQ received at device '%s'", dev2->name);
		  if (dev2->head) {
		    dev2->head->offset = 0;
		    dev2->head->status = 0;
		    if (dev2->head->type != 1)
		      dev2->pktouterr++;
		  }
		  if (dev2->statusout == 2)
		    dev2->statusout = 3;
		  else {
		    /* Just send EOT, then other packets */
		    pack = create_ack(EOT, dev2);
		    if (pack) {
		      dev2->statusout = 1;
		      queue_first(dev2, pack);
		      dev2->num_packs++;
		    }
		  }
		  break;
		
		case ACK:
		  /* Ack for packet. Go to next packet in queue */
		  if (dev2->encrypt) {
		    if ((res - n) < 17)
		      break;
		    if (!decrypt_ack(dev->bytes + offset + n, dev2))
		      break;
		    n = n + 16;
		  }
		  if ((dev2->status == 4) || (dev2->status == 5))
		    send_page(dev2, 1);
		  if (dev2->statusout == 0) {
		    wlog(L_INFO, "ACK received at device '%s'",
			 dev2->name);
		    dequeue_head(dev2);
		    dev2->pktout++;
		  }
		  break;

		case RS:
		  /* Not implemented now. Just as NAK */

		case CAN:
		case NAK:
		  /* NAK or CAN received. Retry packet (or drop) */
		  if (dev2->encrypt) {
		    if ((res - n) < 17)
		      break;
		    if (!decrypt_ack(dev->bytes + offset + n, dev2))
		      break;
		    n = n + 16;
		  }
		  if ((dev2->status == 4) || (dev2->status == 5))
		    send_page(dev2, 1);
		  if (dev2->statusout == 0) {
		    wlog(L_INFO, "%s received at device '%s'",
			 (c == CAN) ? "CAN" : "NAK", dev2->name);
		    dev2->tnri = i_tnri;
		    dev2->cretry--;
		    dev2->pktouterr++;
		    if ((dev2->cretry == 0) || (c == CAN)) {
		      if (dev2->head) {
			wlog(L_WARNING,
			     "Packet dropped (%s) at device '%s'",
			     (c == CAN) ? "CAN" : "too many retries",
			     dev2->name);
			if (dev2->head->encrypt)
			  decrypt_packet(dev2->head->bytes, dev2);
			log_tnpp_packet(err_packs, dev2->head->bytes,
					dev2->transparentcrc,
					(c == CAN) ? "CAN" : "too many NAKs");
			dequeue_head(dev2);
		      }
		    } else
		      if (dev2->head) {
			/* Resend packet (or send next) */
			dev2->statusout = 1;
			dev2->head->status = 1;
			dev2->head->offset = 0;
		      }
		  }
		  break;

		default:
		  /* All others characters */
		  break;
		}
	    }
	    if ((dev2->status != 1) && (dev2->status != 2)
		&& (dev2->status != 3) && (dev2->status != 6)
		&& (dev2->status != 7))
	      dev2->offset = 0;
	    if ((dev2->status == 1)
		&& (dev2->offset >= (MAX_PACKET_SIZE - 1))) {
	      wlog(L_WARNING, "Buffer overflow at device '%s'", dev2->name);
	      dev2->offset = 0;
	      dev2->status = 0;
	      dev2->pktinerr++;
	    }
	  }
	}
      }

      if (FD_ISSET(dev->fd, &wfd)) {
	/* Send some data (for sockets - with sendto) */
	if (((dev->statusout == 2) && (dev->tidle < 0))
	    || (dev->statusout == 1) || (dev->statusout == 3)) {
	  if (dev->statusout == 2) {
	    /* Send ENQ */
	    dev->tidle = i_tidle_enq;
	    dev->cenq--;
	    if (dev->cenq < 0) {
	      wlog(L_WARNING, "No answer at link test on device '%s'",
		   dev->name);
	      dev->cenq = i_cenq;
	      send_page(dev, 0);
	    }
	    wlog(L_ALL, "Trying to send ENQ in device '%s'",
		 dev->name);
	    pack = create_ack(ENQ, dev);
	    if (pack) {
	      queue_first(dev, pack);
	      dev->num_packs++;
	    }
	  } else
	    if (dev->statusout == 3) {
	      /* Send EOT, then ENQ */
	      pack = create_ack(EOT, dev);
	      if (pack) {
		queue_first(dev, pack);
		dev->num_packs++;
	      }
	    }
	  pack = dev->head;
	  if (dev->encrypt && !pack->encrypt)
	    encrypt_packet(pack, dev);
	  len = pack->length - pack->offset;
	  if (len > 0) {
	    wlog(L_ALL, "Send %d bytes in device '%s'",
		 len, dev->name);
	    if (dev->type == 0) {
	      if (dev != ext_dev)
		res = write(dev->fd, &pack->bytes[pack->offset], len);
	      else
		res = sendto(dev->fd, &pack->bytes[pack->offset], len, 0,
			     (struct sockaddr *) &sun_out,
			     strlen(sun_out.sun_path) + 2);
	    } else
	      /* Send UDP-packets via our input device (for right
	       * address and port */
	      res = sendto(in_dev->fd, &pack->bytes[pack->offset], len, 0,
			   (struct sockaddr *) &dev->peer,
			   sizeof(dev->peer));
	    /* Checking for errors */
	    if (res < 0) {
	      /* Errors at write */
	      if ((errno != EINTR) && (errno != EAGAIN)) {
		wlog(L_ERROR, "Error '%s' (%d) at write in device '%s'",
		     strerror(errno), errno, dev->name);
		if ((dev->type == 0) && (dev != ext_dev)) {
		  /* Trying reconnect serial device */
		  reconnect_serial(dev);
		  tnpp_init(dev);
		}
		if (dev != ext_dev)
		  send_page(dev, 2);
	      } else
		wlog(L_ALL, "EINTR or EAGAIN at write in device '%s'",
		     dev->name);
	    } else {
	      wlog(L_ALL, "%d bytes was sent in device '%s'",
		   res, dev->name);
	      dev->tictout = i_tictout;
	      if (len == res) {
		/* Packet end */
		wlog(L_INFO, "Sending of packet done at device '%s'",
		     dev->name);
		if ((pack->length == 1)
		    || ((pack->encrypt) && (pack->length == 17))) {
		  /* Just ACK, NAK or ENQ. Don't wait ack from peer */
		  dequeue_head(dev);
		  if (dev->statusout != 2) {
		    if (dev->statusout != 3) {
		      /* send next packet from queue */
		      if (dev->head) {
			if (dev->head->status != 2) {
			  dev->statusout = 1;
			  dev->head->status = 1;
			} else
			  dev->statusout = 0;
		      } else
			dev->statusout = 0;
		    } else {
		      dev->statusout = 2;
		      dev->tidle = 0;
		    }
		  }
		} else {
		  /* Now must wait ack from peer */
		  if (dev != ext_dev) {
		    pack->status = 2;
		    dev->statusout = 0;
		  } else {
		    dequeue_head(dev);
		    dev->pktout++;
		  }
		}
	      } else
		pack->offset = pack->offset + res;
	    }
	  } else
	    wlog(L_WARNING, "No bytes to send at device '%s'. Bug",
		 dev->name);
	}
      }

      if (FD_ISSET(dev->fd, &efd)) {
	if (errno != 0)
	  wlog(L_WARNING, "Exception with error '%s' (%d) at device '%s'",
	       strerror(errno), errno, dev->name);
      }
    
      if ((t_aft > t_bef) && (dev != in_dev) && (dev != ext_dev)) {
	/* Must decrement timers */
	if ((dev->status == 0) || (dev->statusout == 2))
	  dev->tidle--;
	if ((dev->statusout == 0) && (dev->head))
	  dev->tnri--;
	if ((dev->status == 1) || (dev->status == 2) || (dev->status == 3)
	    || (dev->status == 6) || (dev->status == 7))
	  dev->tict--;
	if (dev->statusout == 1)
	  dev->tictout--;
	if (((dev->tidle < 0) && (dev->statusout != 2)) || (dev->tnri < 0)
	    || (dev->tict < 0) || (dev->tictout < 0)) {
	  /* Time out. Perform link test */ 
	  wlog(L_INFO, "Time out at device '%s'. Link test", dev->name);
	  if (dev->head) {
	    dev->head->offset = 0;
	    dev->head->status = 0;
	  }
	  dev->statusout = 2;
	  dev->status = 4;
	  dev->tidle = -1;
	  dev->tnri = i_tnri;
	  dev->tict = i_tict;
	  dev->tictout = i_tictout;
	}
	if (dev->head)
	  if ((dev->head->bytes[0] == SOH)
	      && ((t_aft - dev->head->begintime.tv_sec) >= timeout_next)) {
	    wlog(L_WARNING, "Timeout on sending packet at device '%s'. Trying to reroute",
		 dev->name);
	    reroute_head(dev);
	    if (dev->head) {
	      dev->head->offset = 0;
	      dev->head->status = 0;
	    }
	    dev->statusout = 2;
	    dev->status = 4;
	    dev->tidle = -1;
	    dev->tnri = i_tnri;
	    dev->tict = i_tict;
	    dev->tictout = i_tictout;
	  }
      }
    }
    t_bef = t_aft;
  }
  wlog(L_WARNING, "Exiting from main loop");
} /* loop */

/* print some characters */
unsigned char *log_chars(FILE *out, unsigned char *data, int len) {
  int i;
  unsigned char *p;
  unsigned char c;

  if (len <= 0)
    return data;
  fprintf(out, " ");
  p = data;
  for(i = 0; i < len; i++) {
    c = *p;
    if ((c >= 0x20) && (c < 0xff))
      fprintf(out, "%c", c);
    else
      fprintf(out, "<0x%02x>", c);
    p++;
  }
  return p;
}

/* Get string hex-value */
int gethex(unsigned char *s, int num) {
  char buf[5];
  int i;

  if ((num > 4) || (num < 1))
    return 0;
  for(i = 0; i < num; i++)
    buf[i] = toupper(s[i]);
  buf[i] = '\0';
  return strtol(buf, NULL, 16);
} /* gethex */

/* log TNPP packet into file */
void log_tnpp_packet(FILE *out, unsigned char *data,
		     int transparentcrc, char *error) {
  int i, k;
  unsigned char *p;
  time_t ltime;
  char buftime[16];
  char *timestamp;
#if defined(USE_MSQL) || defined(USE_POSTGRESQL) || defined(USE_MYSQL)
  char bufquery[5120];
  int dst, src, serial, inertia, crc;
  char *hdate;
  char *htime;
  unsigned char *enddata;
  unsigned char tempc;
  int res;
#ifdef USE_POSTGRESQL
  PGresult *pgres;
#endif
#ifdef USE_MYSQL
  struct tm *tl;
  char bufdate[11];
#endif

  if (use_sql_log) {
    dst = gethex(data + 1, 4);
    inertia = gethex(data + 5, 2);
    src = gethex(data + 7, 4);
    serial = gethex(data + 11, 2);
    for(enddata = (data + 13); ((*enddata != ETX)
				&& (enddata < (data + MAX_PACKET_SIZE - 5)));
	enddata++);
    tempc = *enddata;
    *enddata = '\0';
    if (transparentcrc)
      crc = gethex(enddata + 1, 4);
    else
      crc = enddata[1] + (enddata[2] << 8);
    time(&ltime);
#ifdef USE_MSQL
    strncpy(buftime, msqlUnixTimeToTime(ltime), sizeof(buftime));
    htime = buftime;
    hdate = msqlUnixTimeToDate(ltime);
#else
#ifdef USE_POSTGRESQL
    hdate = ctime(&ltime) + 4;
    memcpy(buftime, hdate + 7, 8);
    buftime[8] = '\0';
    htime = buftime;
#else
#ifdef USE_MYSQL
    tl = localtime(&ltime);
    snprintf(bufdate, sizeof(bufdate) - 1, "%04d-%02d-%02d",
	     tl->tm_year + 1900, tl->tm_mon + 1, tl->tm_mday);
    hdate = bufdate;
    snprintf(buftime, sizeof(buftime) - 1, "%02:%02%:%02",
	     tl->tm_hour, tl->tm_min, tl->tm_sec);
    htime = buftime;
#endif
#endif
#endif
    snprintf(bufquery, sizeof(bufquery) - 1,
	     "INSERT INTO packets VALUES ('%s','%s',%d,%d,%d,%d,'%s',%d,'%s')",
	     hdate, htime,
	     dst, inertia, src, serial,
	     data + 13, crc,
	     error);
    *enddata = tempc;
#ifdef USE_MSQL
    res = msqlQuery(sock_sql, bufquery);
    if (res < 0)
      wlog(L_ERROR, "Error (%s) at execute query", msqlErrMsg);
#else
#ifdef USE_POSTGRESQL
    pgres = PQexec(conn_sql, bufquery);
    if (!pgres)
      wlog(L_ERROR, "Fatal error at execute query");
    PQclear(pgres);
#else
#ifdef USE_MYSQL
    res = mysql_query(conn_sql, bufquery);
    if (res)
      wlog(L_ERROR, "Error (%s) at execute query", mysql_error(conn_sql));
#endif
#endif
#endif
    return;
  }
#endif
  if (!out)
    return;
  time(&ltime);
  timestamp = ctime(&ltime);
  memcpy(buftime, timestamp + 4, 15);
  buftime[15] = '\0';
  if (!one_line_log) {
    i = 0;
    fprintf(out, "\n**** Packet (%s) %s:\n", error, buftime);
    fprintf(out, "%04X: (SOH) %02X\n", i, data[i]);
    i++;
    fprintf(out, "%04X: (dst) ", i);
    for(k = 0; k < 4; k++)
      fprintf(out, "%c", data[i++]);
    fprintf(out, "\n%04X: (inr) ", i);
    for(k = 0; k < 2; k++)
    fprintf(out, "%c", data[i++]);
    fprintf(out, "\n%04X: (src) ", i);
    for(k = 0; k < 4; k++)
      fprintf(out, "%c", data[i++]);
    fprintf(out, "\n%04X: (ser) ", i);
    for(k = 0; k < 2; k++)
      fprintf(out, "%c", data[i++]);
    fprintf(out, "\n** extensions and data:");
    for(k = 0; data[i] != ETX; i++) {
      if ((k & 0x07) == 0)
	fprintf(out, "\n%04X: %02X", i, data[i]);
      else
	fprintf(out, " %02X", data[i]);
      k++;
    }
    fprintf(out, "\n** end of data\n");
    fprintf(out, "%04X: (ETX) %02X\n", i, data[i]);
    i++;
    fprintf(out, "%04X: (CRC) ", i);
    if (transparentcrc) {
      fprintf(out, "(TRANSPARENT) ");
      for(k = 0; k < 4; k++)
	fprintf(out, "%c", data[i++]);
    } else
      fprintf(out, "%02X%02X", data[i + 1], data[i]);
  } else {
    p = data;
    fprintf(out, "%s (%s)", buftime, error);
    p = log_chars(out, p, 1);
    p = log_chars(out, p, 4);
    p = log_chars(out, p, 2);
    p = log_chars(out, p, 4);
    p = log_chars(out, p, 2);
    i = 0;
    while (p[i] != ETX) i++;
    p = log_chars(out, p, i);
    p = log_chars(out, p, 1);
    if (transparentcrc) {
      p = log_chars(out, p, 4);
      fprintf(out, " (TCRC)");
    } else
      fprintf(out, " %02X%02X", p[1], p[0]);
  }
  fprintf(out, "\n");
  fflush(out);
} /* log_tnpp_packet */

/* Close log files (or SQL-connections) */
void log_close() {
  /* Close log-files */
  if (err_packs)
    fclose(err_packs);
  if (log_packs)
    fclose(log_packs);
  if (file_pid)
    unlink(file_pid);
  if (u_sock > -1) {
    close(u_sock);
    unlink(u_sock_name);
  }
#ifdef USE_MSQL
  if (use_sql_log) {
    msqlClose(sock_sql);
    use_sql_log = 0;
  }
#else
#ifdef USE_POSTGRESQL
  if (use_sql_log) {
    PQfinish(conn_sql);
    use_sql_log = 0;
  }
#else
#ifdef USE_MYSQL
  if (use_sql_log) {
    mysql_close(conn_sql);
    use_sql_log = 0;
  }
#endif
#endif
#endif
} /* log_close */

/* Devices and routes config */
tnpp_device curr_dev;
tnpp_route curr_route;
tnpp_route curr_2route;
tnpp_route curr_3route;
int add_device(void *addr, int len, int flag, void *value);
int add_route(void *addr, int len, int flag, void *value);
int add_2route(void *addr, int len, int flag, void *value);
int add_3route(void *addr, int len, int flag, void *value);
int add_default_route(void *addr, int len, int flag, void *value);
#define DEVICE_CONF_NUM 21
conf_rec devices_conf[DEVICE_CONF_NUM] = {
  {PARAM("device"), STRING, NULL,
   0, NULL, 0, add_device},
  {PARAM("devicename"), STRING_NC, (void *) &curr_dev.name,
   0, NULL, 0, std_STRING},
  {PARAM("baud"), DEC_DIGITS, (void *) &curr_dev.serial_speed,
   0, NULL, 0, std_DIGITS},
  {PARAM("clocal"), BOOL_PAR, (void *) &curr_dev.serial_clocal,
   0, NULL, 0, std_BOOL},
  {PARAM("crtscts"), BOOL_PAR, (void *) &curr_dev.serial_crtscts,
   0, NULL, 0, std_BOOL},
  {PARAM("dtrtoggle"), BOOL_PAR, (void *) &curr_dev.serial_dtrtoggle,
   0, NULL, 0, std_BOOL},
  {PARAM("transparentcrc"), BOOL_PAR, (void *) &curr_dev.transparentcrc,
   0, NULL, 0, std_BOOL},
  {PARAM("address"), HEX_DIGITS, (void *) &curr_dev.addr,
   0, NULL, 0, std_DIGITS},
  {PARAM("routefor"), HEX_DIGITS, NULL,
   0, NULL, 0, add_route},
  {PARAM("routelow"), HEX_DIGITS, NULL,
   0, NULL, 0, add_route},
  {PARAM("routehigh"), HEX_DIGITS, (void *) &curr_route.high_addr,
   0, NULL, 0, std_DIGITS},
  {PARAM("route2for"), HEX_DIGITS, NULL,
   0, NULL, 0, add_2route},
  {PARAM("route2low"), HEX_DIGITS, NULL,
   0, NULL, 0, add_2route},
  {PARAM("route2high"), HEX_DIGITS, (void *) &curr_2route.high_addr,
   0, NULL, 0, std_DIGITS},
  {PARAM("route3for"), HEX_DIGITS, NULL,
   0, NULL, 0, add_3route},
  {PARAM("route3low"), HEX_DIGITS, NULL,
   0, NULL, 0, add_3route},
  {PARAM("route3high"), HEX_DIGITS, (void *) &curr_3route.high_addr,
   0, NULL, 0, std_DIGITS},
  {PARAM("blowfishkey"), STRING, (void *) &curr_dev.key_file,
   0, NULL, 0, std_STRING},
  {PARAM("brokenfish"), BOOL_PAR, (void *) &curr_dev.broken_fish,
   0, NULL, 0, std_BOOL},
  {PARAM("message"), STRING_NC, (void *) &curr_dev.page,
   0, NULL, 0, std_STRING},
  {PARAM_NOP("defaultroute"), BOOL_NOP, NULL,
   0, NULL, 0, add_default_route}
};

/* Translation tables config */
tnpp_translate curr_trans;
int add_translate(void *addr, int len, int flag, void *value);
int inccapfunc(void *addr, int len, int flag, void *value);
int deccapfunc(void *addr, int len, int flag, void *value);
#define TRANSLATE_CONF_NUM 5
conf_rec trans_conf[TRANSLATE_CONF_NUM] = {
  {PARAM("translatefrom"), HEX_DIGITS, NULL,
   0, NULL, 0, add_translate},
  {PARAM("translateto"), HEX_DIGITS, (void *) &curr_trans.new_addr,
   0, NULL, 0, std_DIGITS},
  {PARAM("logpacket"), BOOL_PAR, (void *) &curr_trans.logging,
   0, NULL, 0, std_BOOL},
  {PARAM("inccapfuncfor"), STRING, (void *) &curr_trans.cap[0],
   0, NULL, 0, inccapfunc},
  {PARAM("deccapfuncfor"), STRING, (void *) &curr_trans.cap[0],
   0, NULL, 0, deccapfunc}
};

tnpp_device *create_new_dev() {
  tnpp_device *dev;
  int i;

  dev = (tnpp_device *) malloc(sizeof(tnpp_device));
  if (dev) {
    dev->fd = -1;
    dev->type = 0;
    dev->name = NULL;
    dev->device = NULL;
    dev->addr = 1;
    dev->serial_clocal = 0;
    dev->serial_crtscts = 0;
    dev->serial_dtrtoggle = 0;
    dev->transparentcrc = 0;
    memset(&dev->peer, 0, sizeof(struct sockaddr_in));
    dev->bf = NULL;
    dev->encrypt = 0;
    dev->broken_fish = 0;
    dev->key_file = NULL;
    dev->page = NULL;
    dev->page_already = 0;
    dev->head = NULL;
    dev->tail = NULL;
    dev->num_packs = 0;
    dev->status = 255;
    dev->statusout = 255;
    dev->offset = 0;
    dev->tidle = i_tidle;
    dev->tict = i_tict;
    dev->tictout = i_tictout;
    dev->tnri = i_tnri;
    dev->cretry = i_cretry;
    dev->cenq = i_cenq;
    dev->prev = NULL;
    dev->next = NULL;
    for(i = 0; i < SERIAL_TABLE_NUM; i++)
      dev->serials[i] = 0;
    dev->serials_end = 0;
    dev->last_serial = 0;
    dev->pktin = 0;
    dev->pktout = 0;
    dev->pktinerr = 0;
    dev->pktouterr = 0;
  }
  return dev;
} /* create_new_dev */

tnpp_route *create_new_route() {
  tnpp_route *route;

  route = (tnpp_route *) malloc(sizeof(tnpp_route));
  if (route) {
    route->low_addr = 0;
    route->high_addr = 0;
    route->dev = NULL;
    route->next = NULL;
    route->prev = NULL;
  }
  return route;
} /* create_new_route */

tnpp_translate *create_new_trans() {
  tnpp_translate *trans;

  trans = (tnpp_translate *) malloc(sizeof(tnpp_translate));
  if (trans) {
    trans->old_addr = 0;
    trans->new_addr = 0;
    trans->logging = 0;
    trans->next = NULL;
    trans->prev = NULL;
  }
  return trans;
} /* create_new_trans */

void copy_curr_dev() {
  if (tail_dev && curr_dev.device) {
    memcpy(tail_dev, &curr_dev, sizeof(tnpp_device));
    curr_dev.device = NULL;
  }
} /* copy_curr_dev */

void copy_curr_route() {
  if (tail_route && (curr_route.low_addr != -1)) {
    memcpy(tail_route, &curr_route, sizeof(tnpp_route));
    curr_route.low_addr = -1;
  }
} /* copy_curr_route */

void copy_curr_2route() {
  if (tail_2route && (curr_2route.low_addr != -1)) {
    memcpy(tail_2route, &curr_2route, sizeof(tnpp_route));
    curr_2route.low_addr = -1;
  }
} /* copy_curr_2route */

void copy_curr_3route() {
  if (tail_3route && (curr_3route.low_addr != -1)) {
    memcpy(tail_3route, &curr_3route, sizeof(tnpp_route));
    curr_3route.low_addr = -1;
  }
} /* copy_curr_3route */

void copy_curr_trans() {
  if (tail_trans && (curr_trans.old_addr != -1)) {
    memcpy(tail_trans, &curr_trans, sizeof(tnpp_translate));
    curr_trans.old_addr = -1;
  }
} /* copy_curr_trans */

int change_cap_func(char *range, int change) {
  char buf[9];
  int i;
  char *p;

  for (i = 0; i < CAP_TRANS_NUM; i++)
    if (curr_trans.cap[i].change == 0)
      break;
  if (i == CAP_TRANS_NUM)
    return -1;
  p = strchr(range, '-');
  if (!p)
    return -1;
  if ((p - range) > 8)
    return -1;
  memcpy(buf, range, p - range);
  buf[p - range] = '\0';
  curr_trans.cap[i].low = strtoul(buf, NULL, 16);
  if (!*(p + 1))
    return -1;
  curr_trans.cap[i].high = strtoul(p + 1, NULL, 16);
  curr_trans.cap[i].change = change;
  if (i < (CAP_TRANS_NUM - 1))
    curr_trans.cap[i + 1].change = 0;
  return 0;
} /* change_cap_func */

int inccapfunc(void *addr, int len, int flag, void *value) {
  if (change_cap_func((char *) value, 1) < 0)
    return 0;
  else
    return 1;
} /* inccapfunc */

int deccapfunc(void *addr, int len, int flag, void *value) {
  if (change_cap_func((char *) value, -1) < 0)
    return 0;
  else
    return 1;
} /* deccapfunc */

int add_device(void *addr, int len, int flag, void *value) {
  tnpp_device *dev;
  char *s;

  copy_curr_dev();
  copy_curr_route();
  copy_curr_2route();
  copy_curr_3route();
  dev = create_new_dev();
  s = malloc(strlen((char *) value) + 1);
  if (!dev || !s) {
    wlog(L_ERROR, "Malloc failed at create new device record");
    return 0;
  }
  dev->prev = tail_dev;
  if (tail_dev)
    tail_dev->next = dev;
  tail_dev = dev;
  if (!head_dev)
    head_dev = dev;
  strcpy(s, (char *) value);
  dev->name = strdup(s);
  if (!dev->name) {
    wlog(L_ERROR, "Malloc failed at create new device record");
    return 0;
  }
  dev->device = s;
  dev->serial_speed = 9600;
  dev->serial_crtscts = 1;
  memcpy(&curr_dev, dev, sizeof(tnpp_device));
  curr_route.low_addr = -1;
  curr_2route.low_addr = -1;
  curr_3route.low_addr = -1;
  return 1;
} /* add_device */

int add_route(void *addr, int len, int flag, void *value) {
  tnpp_route *route;

  copy_curr_route();
  route = create_new_route();
  if (!route) {
    wlog(L_ERROR, "Malloc failed at create new route record");
    return 0;
  }
  route->low_addr = (int) value;
  route->high_addr = (int) value;
  if (!tail_dev) {
    wlog(L_ERROR, "No device for new route");
    return 0;
  }
  route->dev = tail_dev;
  route->prev = tail_route;
  if (tail_route)
    tail_route->next = route;
  tail_route = route;
  if (!head_route)
    head_route = route;
  memcpy(&curr_route, route, sizeof(tnpp_route));
  return 1;
} /* add_route */

int add_2route(void *addr, int len, int flag, void *value) {
  tnpp_route *route;

  copy_curr_2route();
  route = create_new_route();
  if (!route) {
    wlog(L_ERROR, "Malloc failed at create new route record");
    return 0;
  }
  route->low_addr = (int) value;
  route->high_addr = (int) value;
  if (!tail_dev) {
    wlog(L_ERROR, "No device for new route");
    return 0;
  }
  route->dev = tail_dev;
  route->prev = tail_2route;
  if (tail_2route)
    tail_2route->next = route;
  tail_2route = route;
  if (!head_2route)
    head_2route = route;
  memcpy(&curr_2route, route, sizeof(tnpp_route));
  return 1;
} /* add_route */

int add_3route(void *addr, int len, int flag, void *value) {
  tnpp_route *route;

  copy_curr_3route();
  route = create_new_route();
  if (!route) {
    wlog(L_ERROR, "Malloc failed at create new route record");
    return 0;
  }
  route->low_addr = (int) value;
  route->high_addr = (int) value;
  if (!tail_dev) {
    wlog(L_ERROR, "No device for new route");
    return 0;
  }
  route->dev = tail_dev;
  route->prev = tail_3route;
  if (tail_3route)
    tail_3route->next = route;
  tail_3route = route;
  if (!head_3route)
    head_3route = route;
  memcpy(&curr_3route, route, sizeof(tnpp_route));
  return 1;
} /* add_route */

/* Set NONBLOCK state on descriptor */
void nonblock(int fd) {
  int fl;

  fl = fcntl(fd, F_GETFL);
  fcntl(fd, F_SETFL, fl | O_NONBLOCK);
} /* nonblock */

int add_default_route(void *addr, int len, int flag, void *value) {
  if (!tail_dev) {
    wlog(L_ERROR, "No device for default route");
    return 0;
  }
  default_route = tail_dev;
  return 1;
} /* add_default_route */

int add_translate(void *addr, int len, int flag, void *value) {
  tnpp_translate *trans;

  copy_curr_trans();
  trans = create_new_trans();
  if (!trans) {
    wlog(L_ERROR, "Malloc failed at create new translate record");
    return 0;
  }
  trans->old_addr = (int) value;
  trans->new_addr = (int) value;
  trans->prev = tail_trans;
  trans->cap[0].change = 0;
  if (tail_trans)
    tail_trans->next = trans;
  tail_trans = trans;
  if (!head_trans)
    head_trans = trans;
  memcpy(&curr_trans, trans, sizeof(tnpp_translate));
  return 1;
} /* add_translate */

/* Dump all packets into file */
void dump_packets(tnpp_device *dev) {
  tnpp_packet *pack;

  for(pack = dev->tail; pack; pack = dev->tail) {
    dev->tail = pack->prev;
    if ((pack->bytes[0] == SOH) && !pack->type) {
      wlog(L_INFO, "Packet from device '%s' saved into file '%s'",
	   dev->name, file_queue);
      put_to_file(file_queue, pack->bytes, dev->transparentcrc);
    }
    free(pack);
  }
} /* dump_packets */

/* Read packets from file */
void read_packets() {
  char buf[MAX_PACKET_SIZE];
  int len;
  FILE *f;

  f = fopen(file_queue, "r");
  if (!f) {
    wlog(L_WARNING, "Can't open file '%s' (%s)", file_queue, strerror(errno));
    return;
  }
  while (!feof(f)) {
    if (!fgets(buf, sizeof(buf), f))
      break;
    len = atoi(buf);
    if ((len <= 0) || (len >= MAX_PACKET_SIZE))
      break;
    if (fread(buf, len, 1, f) != 1)
      break;
    route_tnpp_packet(ext_dev, buf, 0);
  }
  fclose(f);
  truncate(file_queue, 0);
} /* read_packets */

/* (re)read devices config (will drop all packets in queues!) */
int reread_device_conf() {
  tnpp_device *dev;
  tnpp_route *route;
  tnpp_packet *pack;
  int res;
  struct hostent *host;
  char buf[128];
  char *p;
  int fkey;

  /* free all routes */
  for(route = tail_route; route; route = tail_route) {
    tail_route = route->prev;
    free(route);
  }
  for(route = tail_2route; route; route = tail_2route) {
    tail_2route = route->prev;
    free(route);
  }
  for(route = tail_3route; route; route = tail_3route) {
    tail_3route = route->prev;
    free(route);
  }
  tail_route = NULL;
  head_route = NULL;
  curr_route.low_addr = -1;
  tail_2route = NULL;
  head_2route = NULL;
  curr_2route.low_addr = -1;
  tail_3route = NULL;
  head_3route = NULL;
  curr_3route.low_addr = -1;

  /* free all packets and devices */
  for(dev = tail_dev; dev; dev = tail_dev) {
    tail_dev = dev->prev;
    dump_packets(dev);
    if (dev->name != dev->device)
      free(dev->name);
    free(dev->device);
    if (dev->fd >= 0)
      close(dev->fd);
    if (dev->page)
      free(dev->page);
    if (dev->key_file) {
      memset(dev->key_file, 0, strlen(dev->key_file));
      free(dev->key_file);
      if (dev->bf) {
	memset((char *) dev->bf, 0, sizeof(tnpp_bf));
	free(dev->bf);
      }
    }
    free(dev);
  }
  head_dev = NULL;
  tail_dev = NULL;
  default_route = NULL;
  curr_dev.device = NULL;

  /* Read devices config */
  res = read_config(file_dev_conf, devices_conf, DEVICE_CONF_NUM, section);
  if (!res) {
    wlog(L_ERROR, "Error in devices config: %s", config_error);
    return 0;
  }
  wlog(L_ALL, "Devices config: %d lines", res);
  copy_curr_dev();
  copy_curr_route();
  copy_curr_2route();
  copy_curr_3route();

  if (!head_dev) {
    wlog(L_ERROR, "No devices found in config");
    return 0;
  }

  /* Initialize all devices */
  flag_fatal_error = 0;
  for(dev = head_dev; dev; dev = dev->next) {
    if (dev->page) {
      p = strchr(dev->page, ':');
      if (!p || !(p = strchr(p, '%')) || (*(p + 1) != 's')
	  || strchr(p + 1, '%')) {
	wlog(L_ERROR, "Error in fault message for device '%s'", dev->name);
	return 0;
      }
      for (p = (strchr(dev->page, ':') + 1); *p != '\0'; p++)
	if (*p == '\\') {
	  switch (*(p + 1)) {
	  case 'r':
	    *p = '\r';
	    break;
	  case 'n':
	    *p = '\n';
	    break;
	  case '\t':
	    *p = '\t';
	    break;
	  case '\\':
	    *p = '\\';
	    break;
	  }
	  memmove(p + 1, p + 2, strlen(p + 2) + 1);
	}
    }
    if (!strncmp("/dev/", dev->device, 5)) {
      /* Serial device */
      dev->type = 0;
      dev->fd = -1;
      reconnect_serial(dev);
      if (flag_fatal_error)
	return 0;
      dev->status = 0;
      dev->statusout = 0;
    } else {
      /* IP-device */
      dev->type = 1;
      dev->fd = socket(AF_INET, SOCK_DGRAM, 0);
      if (dev->fd < 0) {
	wlog(L_ERROR, "Error '%s' (%d) at create new socket",
	     strerror(errno), errno);
	return 0;
      }
      memset((char *) &dev->peer, 0, sizeof(struct sockaddr_in));
      dev->peer.sin_family = AF_INET;
      p = strchr(dev->device, ':');
      buf[sizeof(buf) - 1] = '\0';
      if (p) {
	dev->peer.sin_port = htons(atoi(p+1));
	*p = '\0';
	strncpy(buf, dev->device, sizeof(buf) - 1);
	*p = ':';
      } else {
	dev->peer.sin_port = htons(DEFAULT_PORT);
	strncpy(buf, dev->device, sizeof(buf) - 1);
      }
      /* Resolve host address */
      if ((buf[0] >= '0') && (buf[0] <= '9'))
	dev->peer.sin_addr.s_addr = inet_addr(buf);
      else {
	host = gethostbyname(buf);
	if (!host) {
	  wlog(L_ERROR, "Host address '%s' is unknown", buf);
	  return 0;
	}
	dev->peer.sin_addr.s_addr = *((unsigned long *) host->h_addr);
      }
      res = connect(dev->fd, (struct sockaddr *) &dev->peer,
		    sizeof(dev->peer));
      if (res < 0) {
	wlog(L_ERROR, "Error '%s' (%d) at connect", strerror(errno), errno);
	return 0;
      }
      dev->status = 0;
      dev->statusout = 0;
    }
    if (dev->key_file) {
      if (dev->type != 1) {
	wlog(L_ERROR, "Serial devices cannot use blowfish");
	return 0;
      }
      dev->encrypt = 1;
      dev->bf = (tnpp_bf *) malloc(sizeof(tnpp_bf));
      if (!dev->bf) {
	wlog(L_ERROR, "Malloc failed at create new device");
	return 0;
      }
      fkey = open(dev->key_file, O_RDONLY);
      if (fkey < 0) {
	wlog(L_ERROR, "Error '%s' (%d) at open key file '%s'",
	     strerror(errno), errno, dev->key_file);
	return 0;
      }
      res = read(fkey, buf, 56);
      if (res <= 0) {
	wlog(L_ERROR, "Error at read key file '%s'", dev->key_file);
	return 0;
      }
      close(fkey);
      brokenfish = dev->broken_fish;
      bf_subkeys_init(&dev->bf->skeys, buf, res);
      memset(buf, 0, sizeof(buf));
    }
    tnpp_init(dev);
    send_empty_packet(dev);
    dev->status = 4;
  }

  in_dev = NULL;
  ext_dev = NULL;
  /* Add IP-input device */
  dev = create_new_dev();
  if (!dev) {
    wlog(L_ERROR, "Malloc failed at create new device");
    return 0;
  }
  dev->fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (dev->fd < 0) {
    wlog(L_ERROR, "Error '%s' (%d) at create new socket",
	 strerror(errno), errno);
    return 0;
  }
  res = 1;
  setsockopt(dev->fd, SOL_SOCKET, SO_REUSEADDR, (void *) &res, sizeof(res));
  dev->device = malloc(9);
  if (dev->device)
    strcpy(dev->device, "IP-Input");
  else {
    wlog(L_ERROR, "Malloc failed at create new device");
    return 0;
  }
  dev->name = dev->device;
  dev->status = 0;
  dev->statusout = 255;
  dev->type = 1;
  dev->addr = -1;
  res = bind(dev->fd, (struct sockaddr *) &my_ip_addr, sizeof(my_ip_addr));
  if (res < 0) {
    wlog(L_ERROR, "Error '%s' (%d) at bind", strerror(errno), errno);
    return 0;
  }
  nonblock(dev->fd);
  tail_dev->next = dev;
  dev->prev = tail_dev;
  tail_dev = dev;
  in_dev = dev;

  if (ext_queue || ext_rdonly) {
    dev = create_new_dev();
    if (!dev) {
      wlog(L_ERROR, "Malloc failed at create new device");
      return 0;
    }
    dev->device = malloc(10);
    if (dev->device)
      strcpy(dev->device, "Ext.Queue");
    else {
      wlog(L_ERROR, "Malloc failed at create new device");
      return 0;
    }
    dev->name = dev->device;
    dev->status = 0;
    dev->statusout = 0;
    dev->type = 0;
    dev->addr = -1;
    tail_dev->next = dev;
    dev->prev = tail_dev;
    tail_dev = dev;
    ext_dev = dev;
  }
  return 1;
} /* reread_device_conf */

/* (re)read translation table */
int reread_trans_conf() {
  tnpp_translate *trans;
  int res;

  /* Free all traslation records */
  for(trans = tail_trans; trans; trans = tail_trans) {
    tail_trans = trans->prev;
    free(trans);
  }
  head_trans = NULL;
  tail_trans = NULL;
  curr_trans.old_addr = -1;

  /* Read translation table */
  res = read_config(file_trans_conf, trans_conf, TRANSLATE_CONF_NUM, section);
  if (!res) {
    wlog(L_ERROR, "Error in translation config: %s", config_error);
    return 0;
  }
  wlog(L_ALL, "Translation table: %d lines", res);
  copy_curr_trans();
  return 1;
} /* reread_trans_conf */

/* Main config */
char *ip_addr;
char *file_err_packs;
char *file_trans_packs;
int my_port;
#define MAIN_CONF_NUM 33
conf_rec main_conf[MAIN_CONF_NUM] = {
  {PARAM("bindaddress"), STRING, (void *) &ip_addr,
   0, (void *) "0.0.0.0", 1, std_STRING},
  {PARAM("port"), DEC_DIGITS, (void *) &my_port,
   0, (void *) 4782, 1, std_DIGITS},
  {PARAM("tnppaddress"), HEX_DIGITS, (void *) &my_tnpp_addr,
   0, (void *) 1, 1, std_DIGITS},
  {PARAM("devicefile"), STRING, (void *) &file_dev_conf,
   0, (void *) DEFAULT_DEV_CONFIG, 1, std_STRING},
  {PARAM("savefile"), STRING, (void *) &file_queue,
   0, (void *) DEFAULT_QUEUE_FILE, 1, std_STRING},
  {PARAM("transfile"), STRING, (void *) &file_trans_conf,
   0, (void *) DEFAULT_TRANS_CONFIG, 1, std_STRING},
  {PARAM("badpackets"), STRING, (void *) &file_err_packs,
   0, (void *) "", 1, std_STRING},
  {PARAM("transpackets"), STRING, (void *) &file_trans_packs,
   0, (void *) "", 1, std_STRING},
  {PARAM("pidfile"), STRING, (void *) &file_pid,
   0, (void *) DEFAULT_PID_FILE, 1, std_STRING},
  {PARAM("onelinelog"), BOOL_PAR, (void *) &one_line_log,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("decinertia"), BOOL_PAR, (void *) &flag_dec_inertia,
   0, (void *) 1, 1, std_BOOL},
  {PARAM("tidle"), DEC_DIGITS, (void *) &i_tidle,
   0, (void *) 60, 1, std_DIGITS},
  {PARAM("tnre"), DEC_DIGITS, (void *) &i_tidle_enq,
   0, (void *) 10, 1, std_DIGITS},
  {PARAM("tict"), DEC_DIGITS, (void *) &i_tict,
   0, (void *) 2, 1, std_DIGITS},
  {PARAM("tictout"), DEC_DIGITS, (void *) &i_tictout,
   0, (void *) 2, 1, std_DIGITS},
  {PARAM("tnri"), DEC_DIGITS, (void *) &i_tnri,
   0, (void *) 20, 1, std_DIGITS},
  {PARAM("thold"), DEC_DIGITS, (void *) &i_thold,
   0, (void *) 10, 1, std_DIGITS},
  {PARAM("cretry"), DEC_DIGITS, (void *) &i_cretry,
   0, (void *) 6, 1, std_DIGITS},
  {PARAM("cenq"), DEC_DIGITS, (void *) &i_cenq,
   0, (void *) 6, 1, std_DIGITS},
  {PARAM("chold"), DEC_DIGITS, (void *) &i_chold,
   0, (void *) 24, 1, std_DIGITS},
  {PARAM("sqlserver"), STRING, (void *) &sql_serv,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("sqlport"), STRING, (void *) &sql_port,
   0, (void *) "5432", 1, std_STRING},
  {PARAM("sqldatabase"), STRING, (void *) &sql_base,
   0, (void *) "rtnppd", 1, std_STRING},
  {PARAM("sqluser"), STRING, (void *) &sql_user,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("sqlpassword"), STRING, (void *) &sql_passwd,
   0, (void *) NULL, 1, std_STRING},
  {PARAM("usesqllog"), BOOL_PAR, (void *) &use_sql_log,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("extsocketin"), STRING, (void *) &u_sock_name,
   0, (void *) DEFAULT_SOCKET_IN_FILE, 1, std_STRING},
  {PARAM("extsocketout"), STRING, (void *) &u_sock_name_out,
   0, (void *) DEFAULT_SOCKET_OUT_FILE, 1, std_STRING},
  {PARAM("extqueue"), BOOL_PAR, (void *) &ext_queue,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("extqrdonly"), BOOL_PAR, (void *) &ext_rdonly,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("maxpackets"), DEC_DIGITS, (void *) &max_packs,
   0, (void *) 64, 1, std_DIGITS},
  {PARAM("allowsendall"), BOOL_PAR, (void *) &allowsend,
   0, (void *) 0, 1, std_BOOL},
  {PARAM("timeoutnext"), DEC_DIGITS, (void *) &timeout_next,
   0, (void *) 180, 1, std_DIGITS}
};

/* (re)read all config files */
/* Return 0, if error */
int reread_main_conf() {
  struct hostent *host;
  int res;
  FILE *pid;
  struct sockaddr_un u_addr;

  /* Free some memory */
  if (ip_addr)
    free(ip_addr);
  if (file_err_packs)
    free(file_err_packs);
  if (file_trans_packs)
    free(file_trans_packs);
  if (file_dev_conf)
    free(file_dev_conf);
  if (file_trans_conf)
    free(file_trans_conf);
  if (sql_serv)
    free(sql_serv);
  if (sql_port)
    free(sql_port);
  if (sql_base)
    free(sql_base);
  if (sql_user)
    free(sql_user);
  if (sql_passwd)
    free(sql_passwd);

  log_close();

  if (u_sock_name)
    free(u_sock_name);

  /* read main config */
  res = read_config(file_main_conf, main_conf, MAIN_CONF_NUM, section);
  if (!res) {
    wlog(L_ERROR, "Error in main config: %s", config_error);
    return 0;
  }

  /* open files and resolve name */
  err_packs = NULL;
  log_packs = NULL;
  if (!use_sql_log) {
    if (strlen(file_err_packs) > 0)
      err_packs = fopen(file_err_packs, "a");
    if (strlen(file_trans_packs) > 0)
      log_packs = fopen(file_trans_packs, "a");
  } else {
#ifdef USE_MSQL
    sock_sql = msqlConnect(sql_serv);
    if (sock_sql < 0) {
      wlog(L_ERROR, "Can't connect to SQL-server (%s)", msqlErrMsg);
      use_sql_log = 0;
    } else {
      res = msqlSelectDB(sock_sql, sql_base);
      if (res < 0) {
	wlog(L_ERROR, "Can't select database %s (%s)", sql_base, msqlErrMsg);
	use_sql_log = 0;
      }
    }
#else
#ifdef USE_POSTGRESQL
    conn_sql = PQsetdbLogin(sql_serv, sql_port, NULL, NULL,
			    sql_base, sql_user, sql_passwd);
    if (!conn_sql) {
      wlog(L_ERROR, "Unknown error at connect to SQL-server");
      use_sql_log = 0;
    } else
      if (PQstatus(conn_sql) == CONNECTION_BAD) {
	wlog(L_ERROR, "Can't connect ot SQL-server (%s)",
	     PQerrorMessage(conn_sql));
	use_sql_log = 0;
      }
#else
#ifdef USE_MYSQL
    conn_sql = mysql_real_connect(NULL, sql_serv, sql_user, sql_passwd,
				  sql_base, atoi(sql_port), NULL, 0);
    if (!conn_sql) {
      wlog(L_ERROR, "Unknown error at connect to SQL-server");
      use_sql_log = 0;
    }
#endif
#endif
#endif
  }
  memset((char *) &my_ip_addr, 0, sizeof(struct sockaddr_in));
  my_ip_addr.sin_family = AF_INET;
  my_ip_addr.sin_port = htons(my_port);
  /* Resolve host address */
  if ((ip_addr[0] >= '0') && (ip_addr[0] <= '9'))
    my_ip_addr.sin_addr.s_addr = inet_addr(ip_addr);
  else {
    host = gethostbyname(ip_addr);
    if (!host) {
      wlog(L_ERROR, "Host address '%s' is unknown", ip_addr);
      return 0;
    }
    my_ip_addr.sin_addr.s_addr = *((unsigned long *) host->h_addr);
  }

  /* read devices config and translation table */
  if (!reread_device_conf() || !reread_trans_conf())
    return 0;

  /* Put pid in pidfile */
  pid = fopen(file_pid, "w");
  if (!pid)
    wlog(L_WARNING, "Error %s (%d) at open PID-file",
	 strerror(errno), errno);
  else {
    fprintf(pid, "%d\n", getpid());
    fclose(pid);
  }

  /* Create and bind unix socket for external queue */
  if (ext_queue || ext_rdonly) {
    memset(&u_addr, 0, sizeof(u_addr));
    u_addr.sun_family = AF_UNIX;
    strncpy(u_addr.sun_path, u_sock_name, sizeof(u_addr.sun_path));
    u_sock = socket(AF_UNIX, SOCK_DGRAM, 0);
    if (u_sock < 0) {
      wlog(L_ERROR, "Error %s (%d) at create unix socket",
	   strerror(errno), errno);
      return 0;
    }
    unlink(u_sock_name);
    res = bind(u_sock, (struct sockaddr *) &u_addr,
	       strlen(u_addr.sun_path) + 2);
    if (res < 0) {
      wlog(L_ERROR, "Error %s (%d) at bind unix socket",
	   strerror(errno), errno);
      return 0;
    }
    if (allowsend)
      chmod(u_sock_name, 0622);
    else
      chmod(u_sock_name, 0600);
    ext_dev->fd = u_sock;
    nonblock(u_sock);
    memset(&sun_out, 0, sizeof(sun_out));
    sun_out.sun_family = AF_UNIX;
    strncpy(sun_out.sun_path, u_sock_name_out, sizeof(sun_out.sun_path));
  }

  read_packets();
  /* We ready for run */
  return 1;
}

/* SIGHUP handler (reread all config) */
void sighup(int x) {
  signal(SIGHUP, SIG_IGN);
  flag_reread = 1;
  return;
} /* sighup */

void dump_all_packets() {
  tnpp_device *dev;

  for (dev = head_dev; dev; dev = dev->next)
    dump_packets(dev);
} /* dump_all_packets */

/* SIGTERM handler (save packets and exit) */
void sigterm(int x) {
  dump_all_packets();
  wlog(L_ERROR, "Terminated normally");
  log_close();
  exit(0);
} /* sigterm */

/* Dump device's statistic */
void dump_dev(FILE *out, tnpp_device *dev) {
  int i, j;
  tnpp_packet *pack;
  char *p;
  char buftime[16];
  char *timestamp;
  time_t t;

  fprintf(out, "Device %s (%s)\n", dev->name, dev->device);
  if (dev == ext_dev)
    fprintf(out, "This device is external queue's socket\n");
  if (dev == in_dev)
    fprintf(out, "This device is IP-input socket\n");
  if (dev->encrypt)
    fprintf(out, "The traffic is encrypted\n");
  fprintf(out, "Packets in: %lu, packets out: %lu\n", dev->pktin, dev->pktout);
  fprintf(out, "Errors in: %lu, errors out: %lu\n", dev->pktinerr,
	  dev->pktouterr);
  fprintf(out, "Internal status: %d, %d\n", dev->status, dev->statusout);
  fprintf(out, "Bytes in receiver's buffer: %d\n", dev->offset);
  fprintf(out, "Last serial for this link: %d\n", dev->last_serial);
  fprintf(out, "Number of packets in queue: %d\n", dev->num_packs);
  if ((dev->num_packs > 0) || dev->head) {
    fprintf(out, "Queue listing:\n");
    for (pack = dev->head; pack; pack = pack->next) {
      t = (time_t) pack->begintime.tv_sec;
      timestamp = ctime(&t);
      memcpy(buftime, timestamp + 4, 15);
      buftime[15] = '\0';
      fprintf(out, "%s (Int: %d) Len: %d Off: %d %s:", buftime, pack->status,
	      pack->length, pack->offset, pack->encrypt ? "Enc" : "NoEnc");
      for (j = 0; j < pack->length; j++)
	fprintf(out, " %02X", pack->bytes[j]);
      fprintf(out, " ");
      for (j = 0; j < pack->length; j++)
	if (pack->bytes[j] >= ' ')
	  fprintf(out, "%c", pack->bytes[j]);
	else
	  fprintf(out, ".");
      fprintf(out, "\n");
    }
  }
  fprintf(out, "Tidle %d, Tict %d, Tictout %d, Tnri %d, Cretry %d, Cenq %d\n",
	  dev->tidle, dev->tict, dev->tictout, dev->tnri, dev->cretry,
	  dev->cenq);
#ifdef TIOCMGET
  if ((dev != ext_dev) && (dev->type == 0)) {
    int sers;

    ioctl(dev->fd, TIOCMGET, &sers);
    fprintf(out, "Serial lines: %s %s %s %s %s %s %s %s %s\n",
	    (sers & TIOCM_LE) ? "LE" : "le",
	    (sers & TIOCM_DTR) ? "DTR" : "dtr",
	    (sers & TIOCM_RTS) ? "RTS" : "rts",
	    (sers & TIOCM_CTS) ? "CTS" : "cts",
	    (sers & TIOCM_ST) ? "ST" : "st",
	    (sers & TIOCM_SR) ? "SR" : "sr",
	    (sers & TIOCM_CD) ? "CD" : "cd",
	    (sers & TIOCM_RNG) ? "RNG" : "rng",
	    (sers & TIOCM_DSR) ? "DSR" : "dsr");
  }
#endif
  fprintf(out, "\n");
} /* dump_dev */

/* SIGUSR1 handler (dump statistic to file) */
void sigusr1(int x) {
  time_t t;
  char *ct;
  FILE *out;
  tnpp_device *dev;

  signal(SIGUSR1, sigusr1);
  if (!file_stat)
    return;
  out = fopen(file_stat, "a");
  if (!out) {
    wlog(L_ERROR, "Can't open file '%s' to dump statistic", file_stat);
    return;
  }
  time(&t);
  ct = ctime(&t);
  fprintf(out, "\nrtnppd status at %s", ct);
  for (dev = head_dev; dev; dev = dev->next)
    dump_dev(out, dev);
  fclose(out);
  wlog(L_WARNING, "Statistic dumped into file '%s'", file_stat);
} /* sigusr1 */

void usage() {
  char thishostname[512];

  gethostname(thishostname, 512);
  printf("\nrtnppd v1.7b by Vasim V.\n");
  printf("\nUsage: rtnppd [-ds] [-C CONFIG] [-S SECTION] [-x LOGLEVEL] [-f FILESTAT]\n");
  printf("\nOptions:\n");
  printf("  -d\t\tDo not daemonize\n");
  printf("  -s\t\tUse stderr instead syslog\n");
  printf("  -C CONFIG\tUse config file CONFIG instead '%s'\n",
	 DEFAULT_MAIN_CONFIG);
  printf("  -S SECTION\tUse section [SECTION] in config files instead [%s]\n",
	 thishostname);
  printf("  -x LOGLEVEL\tLog at level LOGLEVEL (default is 2):\n");
  printf("\t\t\t0 - No logging\n");
  printf("\t\t\t1 - Log only errors\n");
  printf("\t\t\t2 - Log errors and warnings\n");
  printf("\t\t\t3 - Log some info about packets, etc\n");
  printf("\t\t\t4 - Log everything\n");
  printf("  -f FILESTAT\tFile for dump statistic into (by SIGUSR1)\n\n");
} /* usage */

/* Ugh. MAIN!!!! */
int main(int argc, char *argv[]) {
  extern char *optarg;
  extern int optind;
  int ch;
  int nodaemon;
  int level;
  int res;
  int levels[5] = {L_FATAL, L_ERROR, L_WARNING, L_INFO, L_ALL};
  struct timeval tp;

  nodaemon = 0;
  log_stderr = 0;
  file_main_conf = DEFAULT_MAIN_CONFIG;
  loglevel = L_WARNING;
  section = NULL;
  file_stat = NULL;

  /* Parse command options */
  while ((ch = getopt(argc, argv, "dsC:x:S:f:")) != -1)
    switch (ch) {
    case 'd':
      nodaemon = 1;
      break;
     case 's':
      log_stderr = 1;
      break;
    case 'C':
      file_main_conf = optarg;
      break;
    case 'x':
      level = atoi(optarg);
      if ((level >= 0) && (level <= 4))
	loglevel = levels[level];
      else {
	usage();
	exit(1);
      }
      break;
    case 'S':
      section = optarg;
      break;
    case 'f':
      file_stat = optarg;
      break;
    case '?':
    default:
      usage();
      exit(1);
    }
  if (!log_stderr)
    openlog("rtnppd", LOG_PID | LOG_NDELAY, LOG_USER);

  /* Clean variables before start */
  ip_addr = NULL;
  file_trans_packs = NULL;
  file_err_packs = NULL;
  file_dev_conf = NULL;
  file_pid = NULL;
  file_trans_conf = NULL;
  err_packs = NULL;
  log_packs = NULL;
  head_dev = NULL;
  tail_dev = NULL;
  default_route = NULL;
  head_route = NULL;
  tail_route = NULL;
  head_trans = NULL;
  tail_trans = NULL;
  in_dev = NULL;
  ext_dev = NULL;
  sql_serv = NULL;
  sql_port = NULL;
  sql_base = NULL;
  sql_user = NULL;
  sql_passwd = NULL;
  u_sock_name = NULL;
  u_sock = -1;
  use_sql_log = 0;
  flag_reread = 0;

  /* daemonize */
  if (!nodaemon) {
    res = fork();
    if (res != 0)
      exit(0);
    setsid();
  }

  do {
    gettimeofday(&tp, NULL);
    srandom(tp.tv_sec + tp.tv_usec);
    
    /* read configs and run */
    if (!reread_main_conf())
      exit(2);

    /* install signal handlers */
    signal(SIGHUP, sighup);
    signal(SIGTERM, sigterm);
    signal(SIGINT, sigterm);
    signal(SIGUSR1, sigusr1);
    
    loop();
    log_close();
  } while (!flag_fatal_error);
  dump_all_packets();
  if (!log_stderr)
    closelog();
  exit(1);
} /* main */
